from __future__ import (
    absolute_import,
    unicode_literals,
)

import abc
import copy
import functools
from typing import (
    Any,
    Callable,
    TypeVar,
)

import six

from pymetrics.instruments import (
    Counter,
    Gauge,
    Histogram,
    Metric,
    Tag,
    Timer,
    TimerResolution,
)


__all__ = (
    'MetricsRecorder',
    'metric_decorator',
)


M = TypeVar('M', bound=Metric)
R = TypeVar('R')


@six.add_metaclass(abc.ABCMeta)
class MetricsRecorder(object):
    @abc.abstractmethod
    def counter(self, name, initial_value=0, **tags):
        # type: (six.text_type, int, **Tag) -> Counter
        """
        Creates a new counter and prepares it for publishing. The initial value is 0 if not specified. Increment the
        counter after it is returned if you do not specify an initial value.

        :param name: The name of the metric
        :param initial_value: The initial value, which defaults to 0
        :param tags: An additional tags you want associated with this metric
        :return: the created counter.
        """

    @abc.abstractmethod
    def histogram(self, name, force_new=False, initial_value=0, **tags):
        # type: (six.text_type, bool, int, **Tag) -> Histogram
        """
        Creates a new histogram for recording arbitrary numbers that can be averaged and summed. The initial value is
        0 if not specified. Set the value after it is returned if you do not specify an initial value.

        :param name: The name of the metric
        :param force_new: Whether to force the creation of a new histogram if there is already an unpublished
                          histogram with the same name and tags.
        :param initial_value: The initial value, which defaults to 0
        :param tags: An additional tags you want associated with this metric
        :return: the created histogram.
        """

    @abc.abstractmethod
    def timer(self, name, force_new=False, resolution=TimerResolution.MILLISECONDS, initial_value=0, **tags):
        # type: (six.text_type, bool, TimerResolution, int, **Tag) -> Timer
        """
        Creates and starts new timer, a special type of histogram that is recorded around the passage of time. The
        initial value is 0 if not specified, and in most cases you should not specify an initial value. The default
        resolution is milliseconds, which is suitable for most use cases. The returned timer will need to be stopped
        before it can be published (or it will be ignored on publication). However, the returned timer is also a
        context manager, so it will stop itself if you surround the code you wish to measure with `with timer(...)`.

        :param name: The name of the metric
        :param force_new: Whether to force the creation of a new timer if there is already an unpublished timer with
                          the same name and tags.
        :param resolution: The resolution at which the timer should record, which defaults to milliseconds
        :param initial_value: The initial value, which defaults to 0
        :param tags: An additional tags you want associated with this metric
        :return: the created and started timer.
        """

    @abc.abstractmethod
    def gauge(self, name, force_new=False, initial_value=0, **tags):
        # type: (six.text_type, bool, int, **Tag) -> Gauge
        """
        Creates a new gauge and prepares it for publishing. The initial value is 0 if not specified. Set the value
        after it is returned if you do not specify an initial value.

        :param name: The name of the metric
        :param force_new: Whether to force the creation of a new gauge if there is already an unpublished gauge with
                          the same name and tags.
        :param initial_value: The initial value, which defaults to 0
        :param tags: An additional tags you want associated with this metric
        :return: the created gauge.
        """

    @abc.abstractmethod
    def publish_all(self):
        # type: () -> None
        """
        Publishes all metrics that have been recorded since the last publish.
        """

    @abc.abstractmethod
    def publish_if_full_or_old(self, max_metrics=18, max_age=10):
        # type: (int, int) -> None
        """
        Publishes all metrics if at least this many metrics have been recorded *or* at least this much time has elapsed
        from the previous publish.

        :param max_metrics: If the recorder is holding at least this many metrics, publish now (defaults to 18, which
                            is a likely-safe amount assuming an MTU of 1500, which is the MTU for Docker containers)
        :param max_age: If the recorder last published at least this many seconds ago, publish now, even if the
                        recorder isn't "full" (isn't holding on to at least `max_metrics`).
        """

    @abc.abstractmethod
    def throttled_publish_all(self, delay=10):
        # type: (int) -> None
        """
        Publishes all metrics that have been recorded since the last publish unless it has been lest than `delay`
        seconds since the last publish.

        :param delay: The minimum number of seconds between publishes
        """

    @abc.abstractmethod
    def clear(self, only_published=False):
        # type: (bool) -> None
        """
        Clear all metrics that have been recorded. However, if `only_published` is `True`, histograms (and timers)
        that have not been published will not be cleared.

        :param only_published: Whether to leave unpublished histograms
        """


def metric_decorator(
    recorder_fetcher,  # type: Callable[[], MetricsRecorder]
    metric_type,  # type: six.text_type
    metric_name,  # type: six.text_type
    *metric_args,  # type: Any
    **metric_kwargs  # type: Any
):
    # type: (...) -> Callable[[Callable[..., R]], Callable[..., R]]
    """
    This decorator can be used on a function or method to provide a shorthand for obtaining the current
    :class:`MetricsRecorder` and using it to record a metric around the function invocation. This is an abstract
    decorator and should be wrapped with your own function to fetch the `MetricsRecorder` in the way you have designed
    for your application.

    Example usage:

    .. code-block:: python

        def timer(name, *args, **kwargs):
            return metrics_decorator(
                some_func_that_returns_a_metrics_recorder,
                'timer',
                name,
                *args,
                **kwargs,
            )

        ...

        @timer('timer_name')
        def some_function(...):
            do_things()

    Multiple metrics can be chained this way, too:

    .. code-block:: python

        @timer('timer_name')
        @timer('other_timer_name')
        @counter('some_counter')
        def some_function(...):
            do_things()

    If `metric_decorator` is called with the `bool` keyword-only argument `include_metric`, the created :class:`Metric`
    instrument will be passed to the wrapped function or method with an extra keyword argument named `metric`.

    :param recorder_fetcher: A callable that returns a :class:`MetricsRecorder` instance
    :param metric_type: A string with value `timer` or `counter` (other instruments currently do not support being
                        called with a decorator)
    :param metric_name: The metric name to use
    :param metric_args: The positional arguments passed to the metric
    :param metric_kwargs: The keyword arguments passed to the metric
    """
    def real_decorator(f):  # type: (Callable[..., R]) -> Callable[..., R]
        @functools.wraps(f)
        def wrapper(*args, **kwargs):  # type: (*Any, **Any) -> R
            m_kwargs = copy.deepcopy(metric_kwargs)
            include_metric = m_kwargs.pop(str('include_metric'), False)

            metric = getattr(recorder_fetcher(), metric_type)(metric_name, *copy.deepcopy(metric_args), **m_kwargs)

            if include_metric:
                kwargs[str('metric')] = metric

            return metric.record_over_function(f, *args, **kwargs)

        return wrapper

    return real_decorator



from __future__ import (
    absolute_import,
    unicode_literals,
)


__version_info__ = (1, 0, 7)
__version__ = '-'.join(filter(None, ['.'.join(map(str, __version_info__[:3])), (__version_info__[3:] or [None])[0]]))



from __future__ import (
    absolute_import,
    unicode_literals,
)

import errno
import logging
import socket
from typing import (
    Iterable,
    List,
    Union,
    cast,
)

from conformity import fields
import six

from pymetrics.instruments import (
    Counter,
    Gauge,
    Histogram,
    Metric,
    Timer,
    TimerResolution,
)
from pymetrics.publishers.base import MetricsPublisher


__all__ = (
    'StatsdPublisher',
)


IP_HEADER_BYTES = 20
UDP_HEADER_BYTES = 8
MAX_IPV4_PACKET_SIZE_BYTES = 65535
MAX_GIG_E_MTU_BYTES = 9000
MAX_FAST_E_MTU_BYTES = 1518
MAX_IPV4_PAYLOAD_SIZE_BYTES = MAX_IPV4_PACKET_SIZE_BYTES - IP_HEADER_BYTES - UDP_HEADER_BYTES
MAX_GIG_E_PAYLOAD_SIZE_BYTES = MAX_GIG_E_MTU_BYTES - IP_HEADER_BYTES - UDP_HEADER_BYTES
MAX_FAST_E_PAYLOAD_SIZE_BYTES = MAX_FAST_E_MTU_BYTES - IP_HEADER_BYTES - UDP_HEADER_BYTES


@fields.ClassConfigurationSchema.provider(fields.Dictionary(
    {
        'host': fields.UnicodeString(description='The host name or IP address on which the Statsd server is listening'),
        'port': fields.Integer(description='The port number on which the Statsd server is listening'),
        'maximum_packet_size': fields.Integer(
            description='The maximum packet size to send (packets will be fragmented above this limit), defaults to '
                        '65000 bytes.',
        ),
        'network_timeout': fields.Any(fields.Float(gt=0.0), fields.Integer(gt=0), description='The network timeout'),
    },
    optional_keys=('maximum_packet_size', 'network_timeout'),
))
class StatsdPublisher(MetricsPublisher):
    """
    A publisher that emits UDP metrics packets to a Statsd consumer over a network connection.

    For Statsd metric type suffixes, see https://github.com/etsy/statsd/blob/master/docs/metric_types.md.
    """

    METRIC_TYPE_COUNTER = b'c'
    METRIC_TYPE_GAUGE = b'g'
    METRIC_TYPE_HISTOGRAM = b'ms'
    METRIC_TYPE_TIMER = b'ms'

    MAXIMUM_PACKET_SIZE = 65000
    """
    Maximum size of a localhost UDP packet, which is used as the maximum maximum and default maximum packet size for
    this publisher (can be configured lower to account for lower MTU).
    """

    def __init__(self, host, port, network_timeout=0.5, maximum_packet_size=MAXIMUM_PACKET_SIZE):
        # type: (six.text_type, int, Union[int, float], int) -> None
        self.host = host
        self.port = port
        self.timeout = network_timeout
        self.maximum_packet_size = min(maximum_packet_size, self.MAXIMUM_PACKET_SIZE)

        self._metric_type_histogram = self.METRIC_TYPE_HISTOGRAM
        self._metric_type_timer = self.METRIC_TYPE_TIMER

    @staticmethod
    def _get_binary_value(string):  # type: (Union[six.text_type, six.binary_type]) -> six.binary_type
        if isinstance(string, six.text_type):
            return string.encode('utf-8')
        return string

    def get_formatted_metrics(self, metrics, enable_meta_metrics=False):
        # type: (Iterable[Metric], bool) -> List[six.binary_type]
        meta_timer = None
        if enable_meta_metrics:
            meta_timer = Timer('', resolution=TimerResolution.MICROSECONDS)

        formatted_metrics = []
        for metric in metrics:
            if metric.value is None:
                continue

            if isinstance(metric, Counter):
                type_label = self.METRIC_TYPE_COUNTER
            elif isinstance(metric, Gauge):
                type_label = self.METRIC_TYPE_GAUGE
            elif isinstance(metric, Timer):
                type_label = self._metric_type_timer
            elif isinstance(metric, Histogram):
                type_label = self._metric_type_histogram
            else:
                continue  # not possible unless a new metric type is added

            formatted_metrics.append(
                b'%s:%d|%s' % (self._get_binary_value(metric.name), metric.value, type_label)
            )

        if not formatted_metrics:
            return []

        if meta_timer:
            meta_timer.stop()
            formatted_metrics.insert(
                0,
                b'pymetrics.meta.publish.statsd.format_metrics:%d|%s' % (
                    cast(int, meta_timer.value),
                    self._metric_type_timer,
                )
            )

        return formatted_metrics

    def publish(self, metrics, error_logger=None, enable_meta_metrics=False):
        # type: (Iterable[Metric], six.text_type, bool) -> None
        if not metrics:
            return

        formatted_metrics = self.get_formatted_metrics(metrics, enable_meta_metrics)
        if not formatted_metrics:
            return

        # The maximum UDP packet size is 65,535 bytes. For localhost, that is also the maximum MTU. For now, all of our
        # metrics are sent over localhost, so we don't worry about the significantly-lower MTU realized over ethernet
        # networks. However, only Statsd allows this maximum size. DogStatsd caps the packet size at 8192. Even at the
        # maximum, we might still exceed that. The loss rate is incredibly tiny, but our analysis has shown that we do,
        # sometimes, lose metrics. So, for now, we chunk the values by the known limits to ensure we stay under those
        # limits.

        chunk = []  # type: List[six.binary_type]
        cumulative_length = 0
        for formatted_metric in formatted_metrics:
            metric_length = len(formatted_metric) + 1  # 1 is the length of a line terminator
            cumulative_length += metric_length

            if cumulative_length > self.maximum_packet_size:
                # This metric would put us over the packet size limit, so send the existing chunk and reset
                self._send_chunked_payload(b'\n'.join(chunk), len(chunk), error_logger, enable_meta_metrics)
                cumulative_length = metric_length
                chunk = []

            chunk.append(formatted_metric)

        if chunk:
            # We have unsent metrics left in the chunk, so send them
            self._send_chunked_payload(b'\n'.join(chunk), len(chunk), error_logger, enable_meta_metrics)

    def _send_chunked_payload(self, payload, number_of_metrics, error_logger=None, enable_meta_metrics=False):
        # type: (six.binary_type, int, six.text_type, bool) -> None
        meta_timer = None
        error = error_max_packet = False
        sock = None
        try:
            if enable_meta_metrics:
                meta_timer = Timer('', resolution=TimerResolution.MICROSECONDS)

            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.settimeout(self.timeout)
            sock.connect((self.host, self.port))
            sock.sendall(payload)
        except Exception as e:
            error = True
            if isinstance(e, socket.error) and e.errno == errno.EMSGSIZE:
                error_max_packet = True

            if error_logger:
                extra = {'data': {
                    'payload_length': len(payload),
                    'num_metrics': number_of_metrics,
                    'enable_meta_metrics': enable_meta_metrics,
                }}
                if error_max_packet:
                    logging.getLogger(error_logger).error(
                        'Failed to send metrics to statsd because UDP packet too big',
                        extra=extra,
                    )
                else:
                    logging.getLogger(error_logger).exception(
                        'Failed to send metrics to statsd {}:{}'.format(self.host, self.port),
                        extra=extra,
                    )
        finally:
            if sock:
                # noinspection PyBroadException
                try:
                    sock.close()
                except Exception:
                    pass
            if meta_timer:
                meta_timer.stop()

        if enable_meta_metrics:
            num_bytes = len(payload)  # TODO temporary; the length of the packet that we tried to send

            payload = b'pymetrics.meta.publish.statsd.send:1|%s' % self.METRIC_TYPE_COUNTER

            payload += b'\npymetrics.meta.publish.statsd.send.num_metrics:%d|%s' % (
                number_of_metrics,
                self._metric_type_histogram
            )

            if meta_timer:
                payload += b'\npymetrics.meta.publish.statsd.send.timer:%d|%s' % (
                    cast(int, meta_timer.value),
                    self._metric_type_timer,
                )

            if error:
                if error_max_packet:
                    payload += b'\npymetrics.meta.publish.statsd.send.error.max_packet:1|%s' % self.METRIC_TYPE_COUNTER
                else:
                    payload += b'\npymetrics.meta.publish.statsd.send.error.unknown:1|%s' % self.METRIC_TYPE_COUNTER

            # TODO The following three stats are temporary, to test out potential MTU problems noted above
            if num_bytes >= MAX_IPV4_PAYLOAD_SIZE_BYTES:
                payload += b'\npymetrics.meta.publish.statsd.send.exceeds_max_packet:1|%s' % self.METRIC_TYPE_COUNTER
            if num_bytes >= MAX_GIG_E_PAYLOAD_SIZE_BYTES:
                payload += b'\npymetrics.meta.publish.statsd.send.exceeds_max_gig_e:1|%s' % self.METRIC_TYPE_COUNTER
            if num_bytes >= MAX_FAST_E_PAYLOAD_SIZE_BYTES:
                payload += b'\npymetrics.meta.publish.statsd.send.exceeds_max_fast_e:1|%s' % self.METRIC_TYPE_COUNTER

            sock = None
            # noinspection PyBroadException
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.settimeout(self.timeout)
                sock.connect((self.host, self.port))
                sock.sendall(payload)
            except Exception:
                if error_logger:
                    logging.getLogger(error_logger).exception(
                        'Failed to send meta metrics to statsd {}:{}'.format(self.host, self.port),
                    )
            finally:
                if sock:
                    # noinspection PyBroadException
                    try:
                        sock.close()
                    except Exception:
                        pass



from __future__ import (
    absolute_import,
    unicode_literals,
)

import enum
import time
from typing import (
    Any,
    Callable,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import six


try:
    from typing import Literal  # type: ignore
except ImportError:
    from typing_extensions import Literal  # type: ignore


__all__ = (
    'Counter',
    'Gauge',
    'Histogram',
    'Metric',
    'Tag',
    'Timer',
    'TimerResolution',
)


R = TypeVar('R')

Tag = Union[six.text_type, six.binary_type, int, float, bool, None]


_valid_initial_values = (int, float)  # type: Tuple[Type, ...]
if six.PY2:
    # noinspection PyUnresolvedReferences
    _valid_initial_values += (long, )  # noqa: F821


class Metric(object):
    """
    A base metric instrument from which all metric instruments inherit. Cannot be instantiated directly.
    """

    def __init__(self, name, initial_value=0, **tags):
        # type: (six.text_type, Union[int, float], **Tag) -> None
        """
        Construct a metric.

        :param name: The metric name
        :param initial_value: The initial value of this metric, which may be an integer or a float (which will be
                              rounded)
        :param tags: The tags associated with this metric (not that not all publishers will support tags)
        """
        if self.__class__ == Metric:
            raise TypeError('Cannot instantiate abstract class "Metric"')
        if not isinstance(name, six.string_types):
            raise TypeError('Metric names must be non-null strings')
        if not isinstance(initial_value, _valid_initial_values):
            raise TypeError('Metric values must be integers or floats')

        self.name = name
        self._initial_value = initial_value
        self._value = self._initial_value
        self.tags = tags

    @property
    def value(self):  # type: () -> Optional[int]
        """
        Returns the value of this metric.

        :return: The metric value
        """
        return int(round(float(self._value)))

    def record_over_function(self, f, *args, **kwargs):  # type: (Callable[..., R], *Any, **Any) -> R
        """
        Records this metric around calling the specified callable with the specified positional and keyword arguments.
        Not all metric types support this. Raises `TypeError` when unsupported.

        :param f: The callable to invoke
        :param args: The positional arguments to pass to the callable
        :param kwargs: The keyword arguments to pass to the callable

        :return: The value the callable returns, unaltered
        """
        raise TypeError('{}s do not support use as a decorator'.format(self.__class__.__name__))

    def __repr__(self):
        return '{}(name="{}", value={})'.format(self.__class__.__name__, self.name, self.value)


class Counter(Metric):
    """
    A counter, for counting the number of times some thing has happened.
    """

    def __init__(self, name, initial_value=0, **tags):  # type: (six.text_type, int, **Tag) -> None
        """
        Construct a counter.

        :param name: The counter name
        :param initial_value: The initial value of this counter, which may be only an integer
        :param tags: The tags associated with this counter (not that not all publishers will support tags)
        """
        if not isinstance(initial_value, six.integer_types) or initial_value < 0:
            raise TypeError('Counter values must be non-null, non-negative integers')

        super(Counter, self).__init__(name, int(initial_value), **tags)

    def increment(self, amount=1):  # type: (int) -> int
        """
        Increments this counter's value by the specified amount.

        :param amount: The amount by which to increment the counter, defaults to 1

        :return: The new value
        """
        self._value += amount
        return self.value

    def reset(self, value=None):  # type: (Optional[int]) -> int
        """
        Resets this counter to the specified value or the initial value if not specified.

        :param value: The value to which to reset this counter, which defaults to the initial value if not specified

        :return: The new value
        """
        if value is None:
            self._value = self._initial_value
        else:
            if value < 0:
                raise ValueError('Counters allow only non-negative integers')
            self._value = int(value)
        return self.value

    @property
    def value(self):  # type: () -> int
        """
        Returns the value of this counter.

        :return: The counter value
        """
        return int(self._value)

    def record_over_function(self, f, *args, **kwargs):  # type: (Callable[..., R], *Any, **Any) -> R
        """
        Increments this counter and then calls the specified callable with the specified positional and keyword
        arguments.

        :param f: The callable to invoke
        :param args: The positional arguments to pass to the callable
        :param kwargs: The keyword arguments to pass to the callable

        :return: The value the callable returns, unaltered
        """
        self.increment()
        return f(*args, **kwargs)


class Histogram(Metric):
    """
    A histogram is a metric for tracking an arbitrary number of something per named activity.
    """

    def set(self, value=None):  # type: (Optional[Union[int, float]]) -> int
        """
        Sets this histogram to the specified value or the initial value if not specified.

        :param value: The value to which to set this histogram, which defaults to the initial value if not specified

        :return: The new value
        """
        if value is None:
            self._value = self._initial_value
        else:
            if value < 0:
                raise ValueError('Histograms allow only non-negative values')
            self._value = value
        return self.value if self.value is not None else 0  # actually not possible to be None here, but satisfy MyPy


class TimerResolution(enum.IntEnum):
    """
    An enum controlling the resolution of published timer values.
    """

    MILLISECONDS = 10**3
    """The timer value will be multiplied by 1,000 and then rounded to an integer before publication"""

    MICROSECONDS = 10**6
    """The timer value will be multiplied by 1,000,000 and then rounded to an integer before publication"""

    NANOSECONDS = 10**9
    """The timer value will be multiplied by 1,000,000,000 and then rounded to an integer before publication"""


class Timer(Histogram):
    """
    A timer is simply a specialized histogram that tracks the arbitrary number of milliseconds, microseconds, or
    nanoseconds per named activity. A single timer instance can be restarted and re-stopped repeatedly, and its value
    will accumulate/increase by the elapsed time each time the timer is stopped. Only if the timer has never been
    started and stopped will the initial or set value be used for publication.
    """

    def __init__(self, name, initial_value=0, resolution=TimerResolution.MILLISECONDS, **tags):
        # type: (six.text_type, Union[int, float], TimerResolution, **Tag) -> None
        """
        Construct a timer.

        :param name: The timer name
        :param initial_value: The initial value of this timer, which may be an integer or a float (which will be
                              rounded)
        :param resolution: The resolution of this timer, which if unset defaults to milliseconds, controls how the
                           value is published (it is multiplied by the resolution factor and then rounded to an
                           integer). It does not affect the initial value, only the value recorded with `start` /
                           `stop` and context manager usage
        :param tags: The tags associated with this timer (not that not all publishers will support tags)
        """
        super(Timer, self).__init__(name, initial_value, **tags)

        self._start_time = None  # type: Optional[float]
        self._running_value = 0.0

        if self._initial_value and self._initial_value > 0:
            self._value = self._initial_value

        self.resolution = resolution

        self.start()

    def start(self):  # type: () -> None
        """
        Starts the timer.
        """
        self._start_time = time.time()

    def stop(self):  # type: () -> None
        """
        Stops the timer.
        """
        if self._start_time is None:
            return  # Cannot stop a timer before it has started
        self._running_value += time.time() - self._start_time
        self._start_time = None

    @property
    def value(self):  # type: () -> Optional[int]
        """
        If the timer has been started and stopped, this returns the total elapsed time of the timer multiplied by the
        resolution and rounded to an integer. Otherwise, this returns the set or initial value of the timer, not
        multiplied, but rounded to an integer.

        :return: The timer value
        """
        if self._running_value > 0 and not self._start_time:
            # If the timer is not currently running but it has previously run, return that amount times the resolution
            # noinspection PyTypeChecker
            return int(round(self._running_value * self.resolution))

        if self._value:
            # Set from initial value, assume the resolution was already correct
            return int(round(float(self._value)))

        return None

    def __enter__(self):  # type: () -> Timer
        """
        Starts the timer at the start of a `with` block.

        :return: `self`
        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # type: (Any, Any, Any) -> Literal[False]
        """
        Stops the timer at the end of a `with` block, regardless of whether an exception occurred.

        :param exc_type: Ignored
        :param exc_value: Ignored
        :param traceback: Ignored

        :return: `False`
        """
        self.stop()
        # noinspection PyTypeChecker
        return False

    def record_over_function(self, f, *args, **kwargs):  # type: (Callable[..., R], *Any, **Any) -> R
        """
        Starts this timer, calls the specified callable with the specified positional and keyword arguments, and then
        stops this timer.

        :param f: The callable to invoke
        :param args: The positional arguments to pass to the callable
        :param kwargs: The keyword arguments to pass to the callable

        :return: The value the callable returns, unaltered
        """
        with self:
            return f(*args, **kwargs)


class Gauge(Metric):
    """
    A gauge is a metric for tracking the ongoing state of something, such as number of items waiting in a queue, size
    of a database or file system, etc.
    """

    def __init__(self, name, initial_value=0, **tags):  # type: (six.text_type, int, **Tag) -> None
        """
        Construct a gauge.

        :param name: The gauge name
        :param initial_value: The initial value of this gauge, which may be an integer
        :param tags: The tags associated with this gauge (not that not all publishers will support tags)
        """
        if not isinstance(initial_value, six.integer_types) or initial_value < 0:
            raise TypeError('Gauge values must be non-null, non-negative integers')

        super(Gauge, self).__init__(name, initial_value, **tags)

    def set(self, value=None):  # type: (Optional[int]) -> int
        """
        Sets this gauge to the specified value or the initial value if not specified.

        :param value: The value to which to set this gauge, which defaults to the initial value if not specified

        :return: The new value
        """
        if value is None:
            self._value = self._initial_value
        else:
            if value < 0:
                raise ValueError('Gauges allow only non-negative integers')
            self._value = value
        return self.value if self.value is not None else 0  # actually not possible to be None here, but satisfy MyPy



from __future__ import (
    absolute_import,
    unicode_literals,
)

from conformity import fields
import six

from pymetrics.instruments import (
    Counter,
    Gauge,
    Histogram,
    Tag,
    Timer,
    TimerResolution,
)
from pymetrics.recorders.base import MetricsRecorder


__all__ = (
    'NonOperationalMetricsRecorder',
    'noop_metrics',
)


@fields.ClassConfigurationSchema.provider(fields.Dictionary(
    {},
    allow_extra_keys=False,
    description='The no-ops recorder has no constructor arguments.',
))
class NonOperationalMetricsRecorder(MetricsRecorder):
    """
    A special metrics recorder that ignores configuration and doesn't keep track of or publish any metrics, useful for
    testing and defaulting a metrics variable to eliminate conditional metrics recording.
    """

    def counter(self, name, initial_value=0, **tags):
        # type: (six.text_type, int, **Tag) -> Counter
        return Counter(name, initial_value, **tags)

    def histogram(self, name, force_new=False, initial_value=0, **tags):
        # type: (six.text_type, bool, int, **Tag) -> Histogram
        return Histogram(name, initial_value, **tags)

    def timer(self, name, force_new=False, resolution=TimerResolution.MILLISECONDS, initial_value=0, **tags):
        # type: (six.text_type, bool, TimerResolution, int, **Tag) -> Timer
        return Timer(name, initial_value, resolution, **tags)

    def gauge(self, name, force_new=False, initial_value=0, **tags):
        # type: (six.text_type, bool, int, **Tag) -> Gauge
        return Gauge(name, initial_value, **tags)

    def publish_all(self):
        # type: () -> None
        """Does nothing"""

    def publish_if_full_or_old(self, max_metrics=18, max_age=10):
        # type: (int, int) -> None
        """Does nothing"""

    def throttled_publish_all(self, delay=10):
        # type: (int) -> None
        """Does nothing"""

    def clear(self, only_published=False):
        # type: (bool) -> None
        """Does nothing"""


noop_metrics = NonOperationalMetricsRecorder()
"""A singleton instance of `NonOperationalMetricsRecorder`."""



from __future__ import (
    absolute_import,
    unicode_literals,
)

from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Union,
    cast,
)

from conformity import fields
import six

from pymetrics.instruments import (
    Counter,
    Gauge,
    Histogram,
    Metric,
    Tag,
    Timer,
    TimerResolution,
)
from pymetrics.publishers.statsd import StatsdPublisher


__all__ = (
    'DogStatsdPublisher',
)


_datadog_tags_value_type = fields.Nullable(
    fields.Any(fields.UnicodeString(), fields.ByteString(), fields.Integer(), fields.Float(), fields.Boolean()),
)


@fields.ClassConfigurationSchema.provider(fields.Dictionary(
    {
        'host': fields.UnicodeString(
            description='The host name or IP address on which the Dogstatsd server is listening',
        ),
        'port': fields.Integer(description='The port number on which the Dogstatsd server is listening'),
        'maximum_packet_size': fields.Integer(
            description='The maximum packet size to send (packets will be fragmented above this limit), defaults to '
                        '8000 bytes.',
        ),
        'network_timeout': fields.Any(fields.Float(gt=0.0), fields.Integer(gt=0), description='The network timeout'),
        'global_tags': fields.SchemalessDictionary(
            key_type=fields.UnicodeString(),
            value_type=_datadog_tags_value_type,
            description='Datadog tags to apply to all published metrics.',
        ),
        'extra_gauge_tags': fields.SchemalessDictionary(
            key_type=fields.UnicodeString(),
            value_type=_datadog_tags_value_type,
            description='Extra datadog tags, in addition to `global_tags` if applicable, to apply to all published '
                        'gauges. This is necessary when multiple processes are simultaneously publishing gauges with '
                        'the same name and you need to create charts or monitors that sum the values of all of these '
                        'gauges across all processes (because Datadog does not support identical distributed gauge '
                        'names+tags and will eliminate duplicates).',
        ),
        'use_distributions': fields.Boolean(
            description='Whether to publish histograms and timers as Datadog distributions. For more information '
                        'about Datadog distributions, see https://docs.datadoghq.com/graphing/metrics/distributions/.'
                        'Defaults to `False`.',
        ),
    },
    optional_keys=('maximum_packet_size', 'network_timeout', 'global_tags', 'extra_gauge_tags', 'use_distributions'),
))
class DogStatsdPublisher(StatsdPublisher):
    """
    A special version of the Statsd publisher than understands the DataDog extensions to the Statsd protocol
    (histograms, distributions, and tags).

    For DogStatsd metric type suffixes, see https://docs.datadoghq.com/developers/dogstatsd/#datagram-format.
    """

    METRIC_TYPE_HISTOGRAM = b'h'
    METRIC_TYPE_DISTRIBUTION = b'd'

    MAXIMUM_PACKET_SIZE = 8000
    """
    Maximum size of the DogStatsd packet buffer, which is used as the maximum and default maximum packet size for this
    publisher (can be configured lower to account for lower MTU). See https://github.com/DataDog/dd-agent/issues/2638
    and https://github.com/DataDog/dd-agent/blob/e805a9a2022803d832cf7f7d8fa8895fd686945c/dogstatsd.py#L367.
    """

    def __init__(
        self,
        host,  # type: six.text_type
        port,  # type: int
        network_timeout=0.5,  # type: Union[int, float]
        global_tags=None,  # type: Dict[six.text_type, Tag]
        extra_gauge_tags=None,  # type: Dict[six.text_type, Tag]
        use_distributions=False,  # type: bool
        maximum_packet_size=MAXIMUM_PACKET_SIZE,  # type: int
    ):
        # type: (...) -> None
        super(DogStatsdPublisher, self).__init__(host, port, network_timeout, maximum_packet_size)

        if global_tags and not isinstance(global_tags, dict):
            raise ValueError('Global tags must be dicts')
        if extra_gauge_tags and not isinstance(extra_gauge_tags, dict):
            raise ValueError('Extra gauge tags must be dicts')

        self._global_tags_string = self._generate_tag_string(global_tags)
        self._global_gauge_tags_string = self._generate_tag_string(extra_gauge_tags, self._global_tags_string)

        if use_distributions:
            self._metric_type_histogram = self.METRIC_TYPE_DISTRIBUTION
            self._metric_type_timer = self.METRIC_TYPE_DISTRIBUTION

    @classmethod
    def _generate_tag_string(
        cls,
        tags,  # type: Optional[Dict[six.text_type, Tag]]
        existing_tags_string=b'',  # type: six.binary_type
    ):
        # type: (...) -> six.binary_type
        if not tags:
            return existing_tags_string

        if existing_tags_string:
            tags_string = existing_tags_string
            first = False
        else:
            tags_string = b'|#'
            first = True

        for tag, value in six.iteritems(tags):
            value_string = b''
            if value is not None:
                if isinstance(value, six.integer_types):
                    value_string = b':%d' % value
                elif isinstance(value, float):
                    value_string = (b':%f' % value).rstrip(b'0')
                else:
                    value_string = b':%s' % cls._get_binary_value(value)
            if first:
                tags_string += b'%s%s' % (cls._get_binary_value(tag), value_string)
                first = False
            else:
                tags_string += b',%s%s' % (cls._get_binary_value(tag), value_string)

        return tags_string

    def get_formatted_metrics(self, metrics, enable_meta_metrics=False):
        # type: (Iterable[Metric], bool) -> List[six.binary_type]
        meta_timer = None
        if enable_meta_metrics:
            meta_timer = Timer('', resolution=TimerResolution.MICROSECONDS)

        formatted_metrics = []
        for metric in metrics:
            if metric.value is None:
                continue

            existing_tags_string = self._global_tags_string

            if isinstance(metric, Counter):
                type_label = self.METRIC_TYPE_COUNTER
            elif isinstance(metric, Gauge):
                type_label = self.METRIC_TYPE_GAUGE
                existing_tags_string = self._global_gauge_tags_string
            elif isinstance(metric, Timer):
                type_label = self._metric_type_timer
            elif isinstance(metric, Histogram):
                type_label = self._metric_type_histogram
            else:
                continue

            metric_tags_string = self._generate_tag_string(metric.tags, existing_tags_string)

            formatted_metrics.append(
                b'%s:%d|%s%s' % (self._get_binary_value(metric.name), metric.value, type_label, metric_tags_string)
            )

        if not formatted_metrics:
            return []

        if meta_timer:
            meta_timer.stop()
            formatted_metrics.insert(
                0,
                b'pymetrics.meta.publish.statsd.format_metrics:%d|%s%s' % (
                    cast(int, meta_timer.value),
                    self._metric_type_timer,
                    self._global_tags_string,
                ),
            )

        return formatted_metrics



from __future__ import (
    absolute_import,
    unicode_literals,
)

from typing import Iterable

from pymetrics.configuration import Configuration
from pymetrics.instruments import Metric


__all__ = (
    'publish_metrics',
)


def publish_metrics(metrics, configuration):  # type: (Iterable[Metric], Configuration) -> None
    """
    Publish a set of metrics via the set of configured publishers.

    :param metrics: The metrics to publish
    :param configuration: The configuration object containing the configured publishers
    """

    for publisher in configuration.publishers:
        publisher.publish(metrics, configuration.error_logger_name, configuration.enable_meta_metrics)



from __future__ import (
    absolute_import,
    unicode_literals,
)

import logging
from typing import (
    Any,
    Iterable,
    Union,
    cast,
)

from conformity import fields
from conformity.fields.logging import PythonLogLevel
import six

from pymetrics.instruments import (
    Counter,
    Gauge,
    Histogram,
    Metric,
    Timer,
)
from pymetrics.publishers.base import MetricsPublisher


__all__ = (
    'LogPublisher',
)


@fields.ClassConfigurationSchema.provider(fields.Dictionary(
    {
        'log_name': fields.UnicodeString(description='The name of the logger to which to publish metrics'),
        'log_level': fields.Any(
            fields.Constant(10, 20, 30, 40, 50),
            PythonLogLevel(),
            description='The log level (name or int) for publishing metrics, defaults to logging.INFO',
        ),
    },
    optional_keys=('log_level', ),
))
class LogPublisher(MetricsPublisher):
    def __init__(self, log_name, log_level=logging.INFO):  # type: (six.text_type, Union[int, six.text_type]) -> None
        self.log_name = log_name
        self.logger = logging.getLogger(self.log_name)

        if isinstance(log_level, int):
            self.log_level = log_level
        else:
            # getLevelName is a misnomer. It returns the name if you pass an int and the int if you pass a name.
            self.log_level = cast(int, logging.getLevelName(log_level))

    @staticmethod
    def _get_str_value(value):  # type: (Any) -> six.text_type
        if isinstance(value, six.binary_type):
            return value.decode('utf-8')
        return six.text_type(value)

    def publish(self, metrics, error_logger=None, enable_meta_metrics=False):
        # type: (Iterable[Metric], six.text_type, bool) -> None
        if not metrics:
            return

        formatted_metrics = []

        for metric in sorted(metrics, key=lambda x: '.'.join((str(type(x)), x.name))):
            if metric.value is None:
                continue

            name = metric.name
            if isinstance(metric, Counter):
                name = '.'.join(('counters', name))
            elif isinstance(metric, Gauge):
                name = '.'.join(('gauges', name))
            elif isinstance(metric, Timer):
                name = '.'.join(('timers', name))
            elif isinstance(metric, Histogram):
                name = '.'.join(('histograms', name))

            if getattr(metric, 'tags', None):
                name += '{{{}}}'.format(
                    ','.join(
                        '{}:{}'.format(k, self._get_str_value(v) if v is not None else '[no value]')
                        for k, v in sorted(metric.tags.items(), key=lambda x: x[0])
                    ),
                )

            formatted_metrics.append(' '.join((name, six.text_type(metric.value))))

        if not formatted_metrics:
            return

        self.logger.log(self.log_level, '; '.join(formatted_metrics))



from __future__ import (
    absolute_import,
    unicode_literals,
)

from conformity import fields

from pymetrics.publishers.base import MetricsPublisher


__all__ = (
    'NullPublisher',
)


@fields.ClassConfigurationSchema.provider(fields.Dictionary({}))
class NullPublisher(MetricsPublisher):
    def publish(self, metrics, error_logger=None, enable_meta_metrics=False):
        """Does nothing"""



from __future__ import (
    absolute_import,
    unicode_literals,
)

import time
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from conformity import fields
from conformity.error import ValidationError
import six

from pymetrics.configuration import (
    CONFIGURATION_SCHEMA,
    Configuration,
    create_configuration,
)
from pymetrics.instruments import (
    Counter,
    Gauge,
    Histogram,
    Metric,
    Tag,
    Timer,
    TimerResolution,
)
from pymetrics.publishers.utils import publish_metrics
from pymetrics.recorders.base import MetricsRecorder


__all__ = (
    'DefaultMetricsRecorder',
)


M = TypeVar('M', bound=Metric)


@fields.ClassConfigurationSchema.provider(fields.Dictionary(
    {
        'prefix': fields.Nullable(fields.UnicodeString(
            description='An optional prefix for all metrics names (the period delimiter will be added for you)',
        )),
        'config': fields.Nullable(CONFIGURATION_SCHEMA),
    },
    allow_extra_keys=False,
    optional_keys=('config', ),
    description='The configuration schema for the default metrics recorder constructor arguments. Without the '
                '`config` key, it will not be able publish metrics.',
))
class DefaultMetricsRecorder(MetricsRecorder):
    def __init__(self, prefix, config=None):
        # type: (Optional[six.text_type], Optional[Dict[six.text_type, Any]]) -> None
        """
        Construct a new recorder.

        :param prefix: A nullable prefix, which if non-null will be prepended to all metrics names, with a single
                       period separating the prefix and the metrics name.
        :param config: The configuration dictionary complying with the official PyMetrics Conformity configuration
                       schema.
        """
        self.prefix = prefix
        self.counters = {}  # type: Dict[six.text_type, Counter]
        self.histograms = {}  # type: Dict[six.text_type, List[Histogram]]
        self.timers = {}  # type: Dict[six.text_type, List[Timer]]
        self.gauges = {}  # type: Dict[six.text_type, List[Gauge]]
        self.unpublished_metrics_count = 0  # type: int
        self._last_publish_timestamp = 0  # type: float

        self._configuration = None  # type: Optional[Configuration]
        self.configure(config)

    @property
    def is_configured(self):  # type: () -> bool
        return self._configuration is not None

    def clear(self, only_published=False):  # type: (bool) -> None
        remaining_histograms = 0
        remaining_timers = 0

        self.counters = {}
        if only_published:
            self.histograms, remaining_histograms = self._get_metric_dict_cleared_of_published_metrics(self.histograms)
            self.timers, remaining_timers = self._get_metric_dict_cleared_of_published_metrics(self.timers)
        else:
            self.histograms = {}
            self.timers = {}
        self.gauges = {}
        self.unpublished_metrics_count = remaining_histograms + remaining_timers

    @staticmethod
    def _get_metric_dict_cleared_of_published_metrics(original):
        # type: (Dict[six.text_type, List[M]]) -> Tuple[Dict[six.text_type, List[M]], int]
        if not original:
            return {}, 0

        new = {}
        remaining = 0
        for name, metrics in six.iteritems(original):
            new_metrics = [metric for metric in metrics if metric.value is None]
            if new_metrics:
                new[name] = new_metrics
                remaining += len(new_metrics)
        return new, remaining

    def _get_name(self, name, other_concerns):
        # type: (six.text_type, Dict[Union[str, six.text_type], Any]) -> Tuple[six.text_type, six.text_type]
        if self.prefix:
            name = '.'.join((self.prefix, name))
        internal_name = name
        if other_concerns and ('resolution' not in other_concerns or len(other_concerns) > 1):
            internal_name += '#{}'.format(str(hash(frozenset(
                (k, v) for k, v in six.iteritems(other_concerns) if k != 'resolution'
            ))))
        return name, internal_name

    def counter(self, name, initial_value=0, **tags):
        # type: (six.text_type, int, **Tag) -> Counter
        name, internal_name = self._get_name(name, tags)

        if internal_name not in self.counters:
            self.counters[internal_name] = Counter(name, initial_value=initial_value, **tags)
            self.unpublished_metrics_count += 1

        return self.counters[internal_name]

    def _get_metric_from_list_or_create(self, collection, name, force_new, metric, initial_value, **kwargs):
        # type: (Dict[six.text_type, List[M]], six.text_type, bool, Type[M], int, **Any) -> M
        name, internal_name = self._get_name(name, kwargs)

        if internal_name not in collection:
            collection[internal_name] = [metric(name, initial_value=initial_value, **kwargs)]
            self.unpublished_metrics_count += 1

        elif force_new or collection[internal_name][-1].value is not None:
            collection[internal_name].append(metric(name, initial_value=initial_value, **kwargs))
            self.unpublished_metrics_count += 1

        return collection[internal_name][-1]

    def histogram(self, name, force_new=False, initial_value=0, **tags):
        # type: (six.text_type, bool, int, **Tag) -> Histogram
        return self._get_metric_from_list_or_create(self.histograms, name, force_new, Histogram, initial_value, **tags)

    def timer(self, name, force_new=False, resolution=TimerResolution.MILLISECONDS, initial_value=0, **tags):
        # type: (six.text_type, bool, TimerResolution, int, **Tag) -> Timer
        return self._get_metric_from_list_or_create(
            self.timers,
            name,
            force_new,
            Timer,
            initial_value,
            resolution=resolution,
            **tags
        )

    def gauge(self, name, force_new=False, initial_value=0, **tags):
        # type: (six.text_type, bool, int, **Tag) -> Gauge
        return self._get_metric_from_list_or_create(self.gauges, name, force_new, Gauge, initial_value, **tags)

    def get_all_metrics(self):  # type: () -> List[Metric]
        meta_timer = None
        if self._configuration and self._configuration.enable_meta_metrics is True:
            meta_timer = Timer('pymetrics.meta.recorder.get_all_metrics', resolution=TimerResolution.MICROSECONDS)

        metrics = []  # type: List[Metric]
        metrics.extend(six.itervalues(self.counters))
        metrics.extend(gauge for gauges in six.itervalues(self.gauges) for gauge in gauges if gauge.value is not None)
        metrics.extend(
            histogram
            for histograms in six.itervalues(self.histograms) for histogram in histograms if histogram.value is not None
        )
        metrics.extend(timer for timers in six.itervalues(self.timers) for timer in timers if timer.value is not None)

        if meta_timer:
            meta_timer.stop()
            metrics.insert(0, meta_timer)

        return metrics

    def configure(self, config=None):  # type: (Optional[Dict[six.text_type, Any]]) -> None
        if not self._configuration:
            # If this recorder is not yet configured
            configuration = None  # type: Optional[Configuration]
            if config:
                # If a local configuration was passed in, use it
                configuration = create_configuration(config)
            else:
                # Attempt to get configuration settings from Django, if available
                config = self.get_config_from_django()
                if config:
                    configuration = create_configuration(config)

            self._configuration = configuration

    # noinspection PyUnresolvedReferences
    @classmethod
    def get_config_from_django(cls):  # type: () -> Optional[Dict[six.text_type, Any]]
        """
        When not in Django context, this is a no-op, returning None. Otherwise, it attempts to return the `METRICS`
        setting from Django settings, if it exists, or the `metrics` setting from the `SOA_SERVER_SETTINGS` setting
        from Django settings, if it exists. `METRICS` is a standard established by this library. `SOA_SERVER_SETTINGS`
        is a standard established in PySOA, which uses PyMetrics.

        :return: The configuration dict from Django settings, if it exists, or None
        """
        settings = cls.get_django_settings()
        if settings:
            if getattr(settings, 'METRICS', None):
                return settings.METRICS
            elif getattr(settings, 'SOA_SERVER_SETTINGS', None):
                return settings.SOA_SERVER_SETTINGS['metrics']['kwargs']['config']

        return None

    class StubImproperlyConfigured(Exception):
        """A stub exception in case Django can't be imported."""

    DjangoImproperlyConfigured = StubImproperlyConfigured
    django_settings = None
    attempted_django_exception_import = False

    # noinspection PyUnresolvedReferences,PyPackageRequirements
    @classmethod
    def get_django_settings(cls):
        if not cls.attempted_django_exception_import:
            # Only attempt to import Django a single time; if it can't be imported once, it'll never import successfully
            cls.attempted_django_exception_import = True
            try:
                from django.core.exceptions import ImproperlyConfigured  # type: ignore
                cls.DjangoImproperlyConfigured = ImproperlyConfigured
            except ImportError:
                pass

        if not cls.django_settings and cls.DjangoImproperlyConfigured is not cls.StubImproperlyConfigured:
            # If the settings haven't been imported yet, but we do have the real Django exception, keep trying to
            # import settings, in case they're improperly configured initially but aren't later.
            try:
                from django.conf import settings  # type: ignore
                if settings:
                    # Django won't actually raise ImproperlyConfigured unless you try to _use_ the settings.
                    getattr(settings, 'DEBUG', False)
                    cls.django_settings = settings
            except (ImportError, cls.DjangoImproperlyConfigured):
                # Could be a circular import or problem with settings that might be resolved at a later time.
                pass
            except ValidationError as e:
                # Likely a circular import that will be resolved at a later time.
                if not (e.args[0] and 'ImportError: ' in e.args[0]):
                    raise

        return cls.django_settings

    def publish_all(self):  # type: () -> None
        if self._configuration:
            metrics = self.get_all_metrics()
            publish_metrics(metrics, self._configuration)
        # silently ignoring situation when not configured, otherwise we'd spam logging
        # clear all regardless if we send, don't want to hog the memory
        self.clear(only_published=True)
        self._last_publish_timestamp = time.time()

    def publish_if_full_or_old(self, max_metrics=18, max_age=10):  # type: (int, int) -> None
        if self.unpublished_metrics_count > max_metrics or time.time() - self._last_publish_timestamp >= max_age:
            self.publish_all()

    def throttled_publish_all(self, delay=10):  # type: (int) -> None
        if time.time() - self._last_publish_timestamp >= delay:
            self.publish_all()



from __future__ import (
    absolute_import,
    unicode_literals,
)

import copy
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import attr
from conformity import (
    fields,
    validator,
)
import six

from pymetrics.publishers.base import MetricsPublisher


__all__ = (
    'Configuration',
    'CONFIGURATION_SCHEMA',
    'create_configuration',
)


CONFIGURATION_SCHEMA = fields.Polymorph(
    switch_field='version',
    contents_map={
        2: fields.Dictionary(
            {
                'version': fields.Constant(2),
                'enable_meta_metrics': fields.Boolean(
                    description='If true, meta-metrics will be recorded documenting the performance of '
                                'PyMetrics itself.',
                ),
                'error_logger_name': fields.UnicodeString(
                    description='By default, errors encountered when publishing metrics are suppressed and lost. If '
                                'this value is truthy, a Logger is created with this name and used to log publication '
                                'errors.',
                ),
                'publishers': fields.Sequence(
                    fields.ClassConfigurationSchema(
                        base_class=MetricsPublisher,
                        description='Import path and arguments for a publisher.',
                    ),
                    min_length=1,
                    description='The configuration for all publishers.',
                ),
            },
            optional_keys=('enable_meta_metrics', 'error_logger_name'),
        ),
    },
    description='The configuration schema changes slightly based on which config version you specify.',
)
""""""  # Empty docstring to make autodoc document this data


@attr.s
class Configuration(object):
    version = attr.ib()  # type: int
    publishers = attr.ib(default=attr.Factory(list))  # type: List[MetricsPublisher]
    error_logger_name = attr.ib(default=None)  # type: Optional[six.text_type]
    enable_meta_metrics = attr.ib(default=False)  # type: bool


@validator.validate_call(
    args=fields.Tuple(copy.deepcopy(CONFIGURATION_SCHEMA)),
    kwargs=None,
    returns=fields.ObjectInstance(Configuration),
)
def create_configuration(config_dict):  # type: (Dict[six.text_type, Any]) -> Configuration
    """
    Creates a `Configuration` object using the provided configuration dictionary. Works in similar fashion to logging's
    configuration.

    Expected format of config is a dict:

    .. code-block:: python

        {
            'version': 2,
            'error_logger_name': 'pymetrics',  # name of the error logger to use, or `None` (the default) to suppress
            'enable_meta_metrics': False,  # whether to enable the collection of meta-metrics
            'publishers': [
                {
                    'path': 'path.to.publisher:ClassName',
                    'kwargs': {
                        ...  # constructor arguments for the publisher
                    },
                },
            ],
        }

    If multiple publishers are specified, metrics will be emitted to each publisher in the order it is specified in
    the configuration list.
    """
    configuration = Configuration(
        version=config_dict['version'],
        enable_meta_metrics=config_dict.get('enable_meta_metrics', False),
        error_logger_name=config_dict.get('error_logger_name'),
    )

    for publisher in config_dict['publishers']:
        configuration.publishers.append(publisher['object'](**publisher.get('kwargs', {})))

    return configuration



from __future__ import (
    absolute_import,
    unicode_literals,
)

import abc
import logging
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
)

import six

from pymetrics.instruments import (
    Counter,
    Gauge,
    Histogram,
    Metric,
    Timer,
)
from pymetrics.publishers.base import MetricsPublisher


__all__ = (
    'SqlPublisher',
)


class SqlPublisher(MetricsPublisher):
    """
    Abstract base class for publishers that publish to SQL databases of any type. Subclasses should implement all the
    backend-specific logic.
    """

    database_type = None  # type: Optional[six.text_type]
    """
    The name of the database backend type, used when logging errors.
    """

    exception_type = Exception  # type: Type[Exception]
    """
    The base class of all possible SQL exceptions this backend could raise, used for catching exceptions in order to
    log them.
    """

    @abc.abstractmethod
    def initialize_if_necessary(self):  # type: () -> None
        """Initialize the database connection, schema, etc., if necessary."""

    @abc.abstractmethod
    def execute_statement_multiple_times(self, statement, arguments):
        # type: (six.text_type, Generator[Tuple[Any, ...], None, None]) -> None
        """
        Execute a given statement multiple times (possibly prepared) using the given generator of argument tuples.

        :param statement: The SQL statement
        :param arguments: A generator of tuples of arguments, one tuple for each time the statement should be executed
        """

    def publish(self, metrics, error_logger=None, enable_meta_metrics=False):
        # type: (Iterable[Metric], six.text_type, bool) -> None
        if not metrics:
            return

        counters = []  # type: List[Counter]
        gauges = {}  # type: Dict[six.text_type, Gauge]
        timers = []  # type: List[Timer]
        histograms = []  # type: List[Histogram]

        for metric in metrics:
            if metric.value is None:
                continue

            if isinstance(metric, Counter):
                counters.append(metric)
            elif isinstance(metric, Gauge):
                gauges[metric.name] = metric
            elif isinstance(metric, Timer):
                timers.append(metric)
            elif isinstance(metric, Histogram):
                histograms.append(metric)

        self.initialize_if_necessary()

        # noinspection PyBroadException
        try:
            if counters:
                self.insert_counters(counters)
        except self.exception_type:
            if error_logger:
                logging.getLogger(error_logger).exception('Failed to send counters to {}'.format(self.database_type))

        # noinspection PyBroadException
        try:
            if gauges:
                self.insert_or_update_gauges(gauges.values())
        except self.exception_type:
            if error_logger:
                logging.getLogger(error_logger).exception('Failed to send gauges to {}'.format(self.database_type))

        # noinspection PyBroadException
        try:
            if timers:
                self.insert_timers(timers)
        except self.exception_type:
            if error_logger:
                logging.getLogger(error_logger).exception('Failed to send timers to {}'.format(self.database_type))

        # noinspection PyBroadException
        try:
            if histograms:
                self.insert_histograms(histograms)
        except self.exception_type:
            if error_logger:
                logging.getLogger(error_logger).exception('Failed to send histograms to {}'.format(self.database_type))

    def insert_counters(self, counters):  # type: (Iterable[Counter]) -> None
        # noinspection SqlNoDataSourceInspection,SqlResolve
        self.execute_statement_multiple_times(
            'INSERT INTO pymetrics_counters (metric_name, metric_value) VALUES (?, ?);',
            ((c.name, c.value) for c in counters),
        )

    def insert_or_update_gauges(self, gauges):  # type: (Iterable[Gauge]) -> None
        # noinspection SqlNoDataSourceInspection,SqlResolve
        self.execute_statement_multiple_times(
            'REPLACE INTO pymetrics_gauges (metric_name, metric_value) VALUES (?, ?);',
            ((g.name, g.value) for g in gauges),
        )

    def insert_timers(self, timers):  # type: (Iterable[Timer]) -> None
        # noinspection SqlNoDataSourceInspection,SqlResolve
        self.execute_statement_multiple_times(
            'INSERT INTO pymetrics_timers (metric_name, metric_value) VALUES (?, ?);',
            ((t.name, (float(t.value) / t.resolution)) for t in timers if t.value is not None),
        )

    def insert_histograms(self, histograms):  # type: (Iterable[Histogram]) -> None
        # noinspection SqlNoDataSourceInspection,SqlResolve
        self.execute_statement_multiple_times(
            'INSERT INTO pymetrics_histograms (metric_name, metric_value) VALUES (?, ?);',
            ((h.name, h.value) for h in histograms),
        )



from __future__ import (
    absolute_import,
    unicode_literals,
)

import contextlib
import sqlite3
from typing import (
    Any,
    Generator,
    Optional,
    Tuple,
    cast,
)

from conformity import fields
import six

from pymetrics.publishers.sql import SqlPublisher


__all__ = (
    'SqlitePublisher',
)


class Sqlite3Connection(sqlite3.Connection):
    """
    An extension to the base connection. The base class is a pure C class on whose instances you can't call setattr.
    This extension enables the use of setattr on connection objects.
    """


@fields.ClassConfigurationSchema.provider(fields.Dictionary(
    {
        'database_name': fields.UnicodeString(description='The name of the Sqlite database to use'),
        'use_uri': fields.Boolean(
            description='Whether the database name should be treated as a URI (Python 3+ only)',
        ),
    },
    optional_keys=('database_name', 'use_uri'),
))
class SqlitePublisher(SqlPublisher):
    """
    A publisher that emits metrics to a Sqlite database file or in-memory database. Especially useful for use in tests
    where you need to actually evaluate your metrics.
    """

    database_type = 'Sqlite'
    exception_type = sqlite3.Error

    MEMORY_DATABASE_NAME = ':memory:'

    _memory_connection = None  # type: Optional[Sqlite3Connection]

    def __init__(self, database_name=MEMORY_DATABASE_NAME, use_uri=False):  # type: (six.text_type, bool) -> None
        if six.PY2 and use_uri:
            raise ValueError('Argument use_uri can only be used in Python 3 and higher')

        self.database_name = database_name
        self.use_uri = use_uri
        self.connection = None  # type: Optional[Sqlite3Connection]

    @staticmethod
    @contextlib.contextmanager
    def connection_context(connection):  # type: (sqlite3.Connection) -> Generator[sqlite3.Cursor, None, None]
        with connection:
            cursor = None
            try:
                cursor = connection.cursor()
                yield cursor
            finally:
                if cursor:
                    cursor.close()

    @contextlib.contextmanager
    def database_context(self):  # type: () -> Generator[sqlite3.Cursor, None, None]
        if not self.connection:
            raise ValueError('Call to database_context before database connection established')
        with self.connection_context(self.connection) as cursor:
            yield cursor

    def initialize_if_necessary(self):  # type: () -> None
        if not self.connection:
            self.connection = self.get_connection(self.database_name, self.use_uri)

        if not getattr(self.connection, '_pymetrics_initialized', None):
            with self.database_context() as cursor:
                # noinspection SqlNoDataSourceInspection,SqlResolve
                cursor.execute("SELECT name FROM sqlite_master WHERE name='pymetrics_counters';")
                needs_schema = cursor.fetchone() is None

            if needs_schema:
                with self.database_context() as cursor:
                    # noinspection SqlNoDataSourceInspection
                    cursor.executescript("""
CREATE TABLE pymetrics_counters (id INTEGER PRIMARY KEY, metric_name TEXT NOT NULL, metric_value INTEGER NOT NULL);

CREATE TABLE pymetrics_gauges (id INTEGER PRIMARY KEY, metric_name TEXT NOT NULL UNIQUE, metric_value INTEGER NOT NULL);

CREATE TABLE pymetrics_timers (id INTEGER PRIMARY KEY, metric_name TEXT NOT NULL, metric_value REAL NOT NULL);

CREATE TABLE pymetrics_histograms (id INTEGER PRIMARY KEY, metric_name TEXT NOT NULL, metric_value INTEGER NOT NULL);
""")

            setattr(self.connection, '_pymetrics_initialized', True)

    @classmethod
    def get_connection(cls, database_name=MEMORY_DATABASE_NAME, use_uri=False):
        # type: (six.text_type, bool) -> Sqlite3Connection
        if six.PY2 and use_uri:
            raise ValueError('Argument use_uri can only be used in Python 3 and higher')

        if database_name == cls.MEMORY_DATABASE_NAME and cls._memory_connection:
            # We only want a single in-memory connection per Python instance
            return cls._memory_connection

        kwargs = {}
        if use_uri:
            kwargs['uri'] = True

        connection = cast(Sqlite3Connection, sqlite3.connect(
            database_name,
            factory=Sqlite3Connection,
            timeout=0.1,
            isolation_level=None,
            check_same_thread=False,
            **kwargs
        ))
        connection.row_factory = sqlite3.Row

        if database_name == cls.MEMORY_DATABASE_NAME:
            cls._memory_connection = connection

        return connection

    def execute_statement_multiple_times(self, statement, arguments):
        # type: (six.text_type, Generator[Tuple[Any, ...], None, None]) -> None
        with self.database_context() as cursor:
            cursor.executemany(statement, arguments)

    @classmethod
    def clear_metrics_from_database(cls, connection):  # type: (sqlite3.Connection) -> None
        with cls.connection_context(connection) as cursor:
            # noinspection SqlNoDataSourceInspection,SqlResolve,SqlWithoutWhere
            cursor.executescript("""
DELETE FROM pymetrics_counters;
DELETE FROM pymetrics_gauges;
DELETE FROM pymetrics_timers;
DELETE FROM pymetrics_histograms;
""")



from __future__ import (
    absolute_import,
    unicode_literals,
)

import abc
from typing import Iterable

import six

from pymetrics.instruments import Metric


__all__ = (
    'MetricsPublisher',
)


@six.add_metaclass(abc.ABCMeta)
class MetricsPublisher(object):
    @abc.abstractmethod
    def publish(self, metrics, error_logger=None, enable_meta_metrics=False):
        # type: (Iterable[Metric], six.text_type, bool) -> None
        """
        Publish the provided metrics in the manner prescribed by the implementation's documentation.

        :param metrics: An iterable of all metrics that should be published
        :param error_logger: The name of the error logger that should be used if an error occurs (if no error logger
                             name is provided, errors will be suppressed)
        :param enable_meta_metrics: If `True`, metrics about the performance of this publisher will also be recorded
                                    (not all publishers will have meta-metrics to record).
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (
    absolute_import,
    unicode_literals,
)

import codecs
import sys

from setuptools import (  # type: ignore
    find_packages,
    setup,
)

from pymetrics import __version__


def readme():
    with codecs.open('README.rst', 'rb', encoding='utf8') as f:
        return f.read()


install_requires = [
    'attrs>=17.4,<22',
    'conformity>=1.26.9,!=1.27.0,<2.0',
    'enum34;python_version<"3.4"',
    'six',
    'typing~=3.7.4;python_version<"3.5"',
    'typing-extensions~=3.10;python_version<"3.8"',
]

mypy_require = [
    'mypy~=0.740;python_version>"3.4"',
    'types-six~=0.1.7;python_version>"3.4"',
    'types-mock~=0.1.3;python_version>"3.4"',
    ]

tests_require = [
    'freezegun',
    'pytest',
    'pytest-cov',
    'pytest-runner',
    'mock',
    'more-itertools~=5.0',  # We must pin this, because 6.0 requires Python 3.
    'importlib-metadata~=5.0;python_version>"3.6"'
] + mypy_require


setup(
    name='pymetrics',
    version=__version__,
    author='Eventbrite, Inc.',
    author_email='opensource@eventbrite.com',
    description='Versatile metrics collection for Python',
    long_description=readme(),
    url='https://github.com/eventbrite/pymetrics',
    packages=list(map(str, find_packages(include=['pymetrics', 'pymetrics.*']))),
    package_data={str('pymetrics'): [str('py.typed')]},  # PEP 561
    zip_safe=False,  # PEP 561
    include_package_data=True,
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=['pytest-runner'] if {'pytest', 'test', 'ptr'}.intersection(sys.argv) else [],
    test_suite='tests',
    extras_require={
        'testing': tests_require,
        'docs': ['sphinx~=2.2;python_version>="3.6"'],
    },
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development',
    ],
    project_urls={
        'Documentation': 'https://pymetrics.readthedocs.io',
        'Issues': 'https://github.com/eventbrite/pymetrics/issues',
        'CI': 'https://travis-ci.org/eventbrite/pymetrics/',
    },
)



from __future__ import (
    absolute_import,
    unicode_literals,
)

from pymetrics.version import (
    __version__,
    __version_info__,
)


__all__ = (
    '__version__',
    '__version_info__',
)
